//
//  GCRSCoding.c
//  NoSynFree
//
//  Created by Liu Chengjian on 17/8/29.
//  Copyright (c) 2017年 csliu. All rights reserved.
//

#include "GCRSCoding.h"
#include "GCRSCommon.h"

void (*gcrs_coders[])(char *dataPtr, char *codeDevPtr,
                      int threadDimX,int blockDimX,
                      int k, int w,
                      int workSizePerGridInLong)
= { gcrs_bandutilize_m_1_coding, gcrs_bandutilize_m_2_coding, gcrs_bandutilize_m_3_coding,gcrs_bandutilize_m_4_coding
};


void (* coding_ptr)(char *dataPtr, char *codeDevPtr,
                    int threadDimX,int blockDimX,
                    int k, int w,
                    int workSizePerGridInLong);

void assign_coding_ptr(int m){
    if (m < MIN_M || m > MAX_M) {
        return;
    }
    
    coding_ptr = gcrs_coders[m - 1];
}

void gcrs_raw_coding_measure(int k, int m, int w,
                             int threadDimX, int blockDimX,
                             int workSizeTotalInLong,
                             char *devDataPtr, char *devCodePtr, float *timeElasped){
    
    
    double timeElapsed, throughputInMB;
    cudaEvent_t startEvent, stopEvent;
    
    assign_coding_ptr(m );
    
    cudaEventCreate(&startEvent);
    cudaEventCreate(&stopEvent);
    
    cudaEventRecord(startEvent,0);
    
    coding_ptr(devDataPtr, devCodePtr,
            threadDimX, blockDimX,
            k,w,
            workSizeTotalInLong);
    
    cudaEventRecord(stopEvent,0);
    cudaThreadSynchronize();
    cudaEventSynchronize(startEvent); //optional
    cudaEventSynchronize(stopEvent); //wait for the event to be executed!
    
    cudaEventElapsedTime(timeElasped, startEvent, stopEvent);
    
    cudaEventDestroy(startEvent);
    cudaEventDestroy(stopEvent);
}
