//
//  GCRSKernel.cu
//  NoSynFree
//
//  Created by Liu Chengjian on 17/8/29.
//  Copyright (c) 2017年 csliu. All rights reserved.
//

#include <cuda_runtime.h>
#include <stdio.h>
#include <math.h>

#include "GCRSCommon.h"

__constant__ unsigned int CONST_DEV_BITMATRIX_CODED_INT[MAX_K*MAX_W*2];

extern "C" cudaError_t gcrs_cuda_sm_set_column_coding_bitmatrix(unsigned int *coding_column_bitmatrix,int k,int m,int w, int factor){
    return cudaMemcpyToSymbol(CONST_DEV_BITMATRIX_CODED_INT, coding_column_bitmatrix, k*w*factor*sizeof(unsigned int), 0, cudaMemcpyHostToDevice);
}

__global__ void gcrs_bandutilize_m_1_coding_dotprod(long *in, long *out, int k, int w, int size){
    extern __shared__ long shared_data[];
    int blockUnits = blockDim.x / w;
    int blockPackets = blockUnits * w;
    int tid = threadIdx.x +blockIdx.x * blockPackets;
    int unit_id_offset = threadIdx.x / w * w;
    int unit_inner_id = tid % w;
    int i,j;

    if (threadIdx.x >= blockPackets){
        return;
    }

    if (tid >= size )
        return;

    long result = 0;

    // row for each thread in the bitmatrix * row size which is k * w
    int index = 0;
    
    unsigned int bitInt = 0x01;
    unsigned int matrixInt;
    
    for ( i = 0; i < k; i++ ) {
            shared_data[threadIdx.x] = *(in + size * i + tid);
            __syncthreads();

        for ( j = 0; j < w; j++ ) {

            matrixInt = CONST_DEV_BITMATRIX_CODED_INT[index];
            if (((matrixInt & (bitInt << unit_inner_id)) >> unit_inner_id) == bitInt){
                result = result^shared_data[unit_id_offset+j];
            }
            ++index;
        }
        __syncthreads();

    }

    *(out + tid) = result;
}

__global__ void gcrs_bandutilize_m_2_coding_dotprod(long *in, long *out, int k, int w, int size){
    extern __shared__ long shared_data[];
    int blockUnits = blockDim.x / w;
    int blockPackets = blockUnits * w;
    int tid = threadIdx.x +blockIdx.x * blockPackets;
    int unit_id_offset = threadIdx.x / w * w;
    int unit_inner_id = tid % w;
    int i,j;
    
    if (threadIdx.x >= blockPackets){
        return;
    }
    
    if (tid >= size )
        return;
    
    long result[2];
    
    result[0]= 0;
    result[1]= 0;
    
    // row for each thread in the bitmatrix * row size which is k * w
    int index = 0;
    
    unsigned int bitInt = 0x01;
    unsigned int matrixInt;
    
    for ( i = 0; i < k; i++ ) {
        shared_data[threadIdx.x] = *(in + size * i + tid);
        __syncthreads();
        
        for ( j = 0; j < w; j++ ) {
            
            matrixInt = CONST_DEV_BITMATRIX_CODED_INT[index];
            
            if (((matrixInt & (bitInt << unit_inner_id)) >> unit_inner_id) == bitInt){
                result[0] = result[0]^shared_data[unit_id_offset+j];
            }
            
            int shiftOffset = (unit_inner_id  + w);
            if (((matrixInt & (bitInt << shiftOffset)) >> shiftOffset) == bitInt){
                result[1] = result[1]^shared_data[unit_id_offset+j];
            }
            
            ++index;
        }
        __syncthreads();
        
    }
    
    *(out + tid) = result[0];
    *(out + size + tid) = result[1];
}

__global__ void gcrs_bandutilize_m_3_coding_dotprod(long *in, long *out, int k, int w, int size){
    extern __shared__ long shared_data[];
    int blockUnits = blockDim.x / w;
    int blockPackets = blockUnits * w;
    int tid = threadIdx.x +blockIdx.x * blockPackets;
    int unit_id_offset = threadIdx.x / w * w;
    int unit_inner_id = tid % w;
    int i,j;
    
    if (threadIdx.x >= blockPackets){
        return;
    }
    
    if (tid >= size )
        return;
    
    long result[3];
    
    result[0]= 0;
    result[1]= 0;
    result[2]= 0;
    
    // row for each thread in the bitmatrix * row size which is k * w
    int index = 0;
    
    unsigned int bitInt = 0x01;
    unsigned int matrixInt;
    
    for ( i = 0; i < k; i++ ) {
        shared_data[threadIdx.x] = *(in + size * i + tid);
        __syncthreads();
        
        for ( j = 0; j < w; j++ ) {
            
            matrixInt = CONST_DEV_BITMATRIX_CODED_INT[index];
            
            if (((matrixInt & (bitInt << unit_inner_id)) >> unit_inner_id) == bitInt){
                result[0] = result[0]^shared_data[unit_id_offset+j];
            }

            int shiftOffset = (unit_inner_id  + w);
            if (((matrixInt & (bitInt << shiftOffset)) >> shiftOffset) == bitInt){
                result[1] = result[1]^shared_data[unit_id_offset+j];
            }
            
            shiftOffset = (unit_inner_id  + 2*w);
            if (((matrixInt & (bitInt << shiftOffset)) >> shiftOffset) == bitInt){
                result[2] = result[2]^shared_data[unit_id_offset+j];
            }

            ++index;
        }
        __syncthreads();
        
    }
    
    *(out + tid) = result[0];
    *(out + size + tid) = result[1];
    *(out + 2*size + tid) = result[2];
}

__global__ void gcrs_bandutilize_m_4_coding_dotprod(long *in, long *out, int k, int w, int size){
    extern __shared__ long shared_data[];
    int blockUnits = blockDim.x / w;
    int blockPackets = blockUnits * w;
    int tid = threadIdx.x +blockIdx.x * blockPackets;
    int unit_id_offset = threadIdx.x / w * w;
    int unit_inner_id = tid % w;
    int i,j;
    
    if (threadIdx.x >= blockPackets){
        return;
    }
    
    if (tid >= size )
        return;
    
    long result[4];
    
    result[0]= 0;
    result[1]= 0;
    result[2]= 0;
    result[3]= 0;
    
    // row for each thread in the bitmatrix * row size which is k * w
    int index = 0;
    
    unsigned int bitInt = 0x01;
    unsigned int matrixInt;
    
    for ( i = 0; i < k; i++ ) {
        shared_data[threadIdx.x] = *(in + size * i + tid);
        __syncthreads();
        
        for ( j = 0; j < w; j++ ) {
            
            matrixInt = CONST_DEV_BITMATRIX_CODED_INT[index];
            
            if (((matrixInt & (bitInt << unit_inner_id)) >> unit_inner_id) == bitInt){
                result[0] = result[0]^shared_data[unit_id_offset+j];
            }
            
            int shiftOffset = (unit_inner_id  + w);
            if (((matrixInt & (bitInt << shiftOffset)) >> shiftOffset) == bitInt){
                result[1] = result[1]^shared_data[unit_id_offset+j];
            }
            
            shiftOffset = (unit_inner_id  + 2*w);
            if (((matrixInt & (bitInt << shiftOffset)) >> shiftOffset) == bitInt){
                result[2] = result[2]^shared_data[unit_id_offset+j];
            }
            
            shiftOffset = (unit_inner_id  + 3*w);
            if (((matrixInt & (bitInt << shiftOffset)) >> shiftOffset) == bitInt){
                result[3] = result[3]^shared_data[unit_id_offset+j];
            }
            
            ++index;
        }
        __syncthreads();
        
    }
    
    *(out + tid) = result[0];
    *(out + size + tid) = result[1];
    *(out + 2*size + tid) = result[2];
    *(out + 3*size + tid) = result[3];

}

extern "C" void gcrs_bandutilize_m_1_coding(char *dataPtr, char *codeDevPtr,
                                            int threadDimX,int blockDimX,
                                            int k, int w,
                                            int workSizePerGridInLong){
    dim3 gridDim(blockDimX, 1, 1);
    dim3 blockDim(threadDimX, 1, 1);
    
    gcrs_bandutilize_m_1_coding_dotprod<<<gridDim, blockDim, threadDimX*sizeof(long)>>>( (long *)dataPtr, (long *)codeDevPtr,k, w, workSizePerGridInLong);
    
}

extern "C" void gcrs_bandutilize_m_2_coding(char *dataPtr, char *codeDevPtr,
                                            int threadDimX,int blockDimX,
                                            int k, int w,
                                            int workSizePerGridInLong){
    dim3 gridDim(blockDimX, 1, 1);
    dim3 blockDim(threadDimX, 1, 1);
    
    gcrs_bandutilize_m_2_coding_dotprod<<<gridDim, blockDim, threadDimX*sizeof(long)>>>( (long *)dataPtr, (long *)codeDevPtr,k, w, workSizePerGridInLong);
    
}

extern "C" void gcrs_bandutilize_m_3_coding(char *dataPtr, char *codeDevPtr,
                                            int threadDimX,int blockDimX,
                                            int k, int w,
                                            int workSizePerGridInLong){
    dim3 gridDim(blockDimX, 1, 1);
    dim3 blockDim(threadDimX, 1, 1);
    
    gcrs_bandutilize_m_3_coding_dotprod<<<gridDim, blockDim, threadDimX*sizeof(long)>>>( (long *)dataPtr, (long *)codeDevPtr,k, w, workSizePerGridInLong);
    
}

extern "C" void gcrs_bandutilize_m_4_coding(char *dataPtr, char *codeDevPtr,
                                            int threadDimX,int blockDimX,
                                            int k, int w,
                                            int workSizePerGridInLong){
    dim3 gridDim(blockDimX, 1, 1);
    dim3 blockDim(threadDimX, 1, 1);
    
    gcrs_bandutilize_m_4_coding_dotprod<<<gridDim, blockDim, threadDimX*sizeof(long)>>>( (long *)dataPtr, (long *)codeDevPtr,k, w, workSizePerGridInLong);
    
}

