//
//  GCRSKernel.cu
//  NoSynFree
//
//  Created by Liu Chengjian on 17/8/29.
//  Copyright (c) 2017年 csliu. All rights reserved.
//

#include <cuda_runtime.h>
#include <stdio.h>
#include <math.h>

#include "GCRSCommon.h"

__constant__ unsigned int CONST_DEV_BITMATRIX_CODED_INT[MAX_K*MAX_W*2];

extern "C" cudaError_t gcrs_cuda_sm_set_column_coding_bitmatrix(unsigned int *coding_column_bitmatrix,int k,int m,int w, int factor){
    return cudaMemcpyToSymbol(CONST_DEV_BITMATRIX_CODED_INT, coding_column_bitmatrix, k*w*factor*sizeof(unsigned int), 0, cudaMemcpyHostToDevice);
}

__global__ void gcrs_bitmatrixopt_coding_dotprod(long *in, long *out, int k, int w, int size){
    extern __shared__ long shared_data[];
    int blockUnits = blockDim.x / w;
    int blockPackets = blockUnits * w;
    int tid = threadIdx.x +blockIdx.x * blockPackets;
    int unit_id_offset = threadIdx.x / w * w;
    int unit_inner_id = tid % w;
    int i,j;

    if (threadIdx.x >= blockPackets){
        return;
    }

    if (tid >= size )
        return;

    long result = 0;

    // row for each thread in the bitmatrix * row size which is k * w
    int index = 0;
    
    unsigned int bitInt = 0x01;
    unsigned int matrixInt;
    
    for ( i = 0; i < k; i++ ) {
        if(threadIdx.y == 0)
            shared_data[threadIdx.x] = *(in + size * i + tid);
            __syncthreads();

        for ( j = 0; j < w; j++ ) {

            matrixInt = CONST_DEV_BITMATRIX_CODED_INT[index];
            int shiftOffset = (unit_inner_id  + w * threadIdx.y);
            if (((matrixInt & (bitInt << shiftOffset)) >> shiftOffset) == bitInt){
                result = result^shared_data[unit_id_offset+j];
            }
            ++index;
        }
        __syncthreads();

    }

    *(out + size * threadIdx.y + tid) = result;
}

extern "C" void gcrs_bitmatrixopt_coding(char *dataPtr, char *codeDevPtr,
                                        int threadDimX,int blockDimX,
                                        int k, int m, int w,
                                        int workSizePerGridInLong){
    dim3 gridDim(blockDimX, 1, 1);
    dim3 blockDim(threadDimX, m, 1);
    
    gcrs_bitmatrixopt_coding_dotprod<<<gridDim, blockDim, threadDimX*sizeof(long)>>>( (long *)dataPtr, (long *)codeDevPtr,k, w, workSizePerGridInLong);
    
}
