//
//  PErasureBaseKernel.cu
//  PBase
//
//  Created by Liu Chengjian on 16/10/27.
//  Copyright (c) 2016年 csliu. All rights reserved.
//

#include <cuda_runtime.h>
#include <stdio.h>
#include <math.h>

__global__ void crs_coding_dotprod(long *in, long *out, int *bitmatrix, int k, int w, int size){
    
    int blockUnits = blockDim.x / w;
    int blockPackets = blockUnits * w;
    int tid = threadIdx.x +blockIdx.x * blockPackets;
    int unit_id_offset = tid / w * w;
    int unit_inner_id = tid % w;
    int i,j;
    
    if (threadIdx.x >= blockPackets){
        return;
    }
    
    if (tid >= size )
        return;
    
    int index = threadIdx.y*k*w*w + unit_inner_id*k*w;
    
    long result = 0;
    
    for ( i = 0; i < k ; ++i) {
        for (j = 0; j < w ; ++j) {
            
            if (bitmatrix[index] == 1) {
                long input = *(in + size * i +unit_id_offset + j);
                result = result^input;
            }
            
            ++index;
        }
    }
    
    *(out + size * threadIdx.y + tid) = result;
}

extern "C" void perform_base_coding(char *in, char *out,
                                    int *bitmatrix,
                                    int k, int m, int w,
                                    int threadDimX,int blockDimX,
                                    int workSizePerGridInLong){
    dim3 gridDim(blockDimX, 1, 1);
    dim3 blockDim(threadDimX, m, 1);
//    printf("k:%d,m:%d,w:%d,threadDimX:%d,blockDimX:%d,workSizePerGridInLong:%d\n",k,m,w,threadDimX,blockDimX,workSizePerGridInLong);
    
    crs_coding_dotprod<<<gridDim, blockDim>>>((long *)in, (long *)out, bitmatrix, k, w, workSizePerGridInLong);
}