//
//  PErasureMeasure.c
//  PBase
//
//  Created by Liu Chengjian on 16/10/27.
//  Copyright (c) 2016年 csliu. All rights reserved.
//

#include "PErasureMeasure.h"


#include <cuda_runtime.h>
#include <time.h>
#include <math.h>

#include "PErasureCommon.h"

#include "PErasureMatrix.h"
#include "PErasureUtilities.h"

#include "jerasure.h"

float measure_coding(char *in, char *out,
                     int *bitmatrix,
                     int k, int m, int w,
                     int threadDimX,int blockDimX,
                     int workSizePerGridInLong){
    cudaEvent_t startEvent, stopEvent;
    
    float elapsedTime;
    
    cudaEventCreate(&startEvent);
    cudaEventCreate(&stopEvent);
    
    cudaEventRecord(startEvent, 0);

    perform_base_coding(in, out, bitmatrix, k, m, w, threadDimX, blockDimX, workSizePerGridInLong);

    cudaEventRecord(stopEvent,0);
    cudaThreadSynchronize();
    cudaEventSynchronize(startEvent); //optional
    cudaEventSynchronize(stopEvent); //wait for the event to be executed!
    
    cudaEventElapsedTime(&elapsedTime, startEvent, stopEvent);
    
    cudaEventDestroy(startEvent);
    cudaEventDestroy(stopEvent);
    
//    printf("elapsedTime:%fms\n",elapsedTime);
    
    return elapsedTime;
}

void measure_correctness(int k, int m, int w){
    w = 8;

    char *dataHost, *codeHost, **dataPtrs, **codePtrs;
    char *inHost, *outHost, *in, *out ;
    
    int *bitmatrix, *devBitmatrix;
    int threadDimx = THREADS_PER_BLOCK;
    int packetsPerBlock = threadDimx/w *w;
    int blockNum;
    int workSizeInLong;

    int kIdx, mIdx;
    int workSizePerDataParityBlock = 1024*1024;
    
    
    bitmatrix = perasure_create_bitmatrix(k, m, w);
//    verifyBitmatrix = perasure_create_bitmatrix(k, m, w);
    workSizePerDataParityBlock =  (int)align_value(workSizePerDataParityBlock, sizeof(long)*w);
    workSizeInLong = workSizePerDataParityBlock / sizeof(long);
    blockNum = (workSizeInLong/packetsPerBlock);
    
    if ((workSizeInLong%packetsPerBlock) != 0) {
        ++blockNum;
    }
    
    dataHost = talloc(char, workSizePerDataParityBlock*k);
    codeHost = talloc(char, workSizePerDataParityBlock*m);
    
    inHost = talloc(char, workSizePerDataParityBlock*k);
    outHost = talloc(char, workSizePerDataParityBlock*m);
    
    generateRandomValue(dataHost, workSizePerDataParityBlock*k);
    
    dataPtrs = talloc(char **, k);
    codePtrs = talloc(char **, m);
    
    for (kIdx = 0; kIdx < k; ++kIdx) {
        if (kIdx < m) {
            codePtrs[kIdx] = codeHost + workSizePerDataParityBlock * kIdx;
        }
        dataPtrs[kIdx] = dataHost + workSizePerDataParityBlock * kIdx;
    }

    alloc_cuda_device_memory((void **)&devBitmatrix, sizeof(int)*k*w*m*w);
    
    alloc_cuda_device_memory((void **)&in, workSizePerDataParityBlock*k);
    alloc_cuda_device_memory((void **)&out, workSizePerDataParityBlock*m);
    
    
    memcpy(inHost, dataHost, workSizePerDataParityBlock*k);
    
    cudaMemcpy(in, inHost, workSizePerDataParityBlock*k, cudaMemcpyHostToDevice);

    cudaMemcpy(devBitmatrix, bitmatrix, sizeof(int)*k*m*w*w, cudaMemcpyHostToDevice);
//    cudaMemcpy(verifyBitmatrix, devBitmatrix, sizeof(int)*k*m*w*w, cudaMemcpyDeviceToHost);
    
//    printMatrix(bitmatrix, k*w, m*w);
//    printf("\n*************\n");
//    printMatrix(verifyBitmatrix, k*w, m*w);
    
    jerasure_bitmatrix_encode(k, m, w, bitmatrix, dataPtrs, codePtrs, workSizePerDataParityBlock, sizeof(long));

    measure_coding(in, out, devBitmatrix, k, m, w, threadDimx, blockNum, workSizeInLong);
    
    cudaMemcpy(outHost, out,  workSizePerDataParityBlock*m, cudaMemcpyDeviceToHost);
    
    //verify;
    
    int verifyIdx =0;
    for (mIdx = 0; mIdx < m; ++mIdx) {
        for (verifyIdx = 0; verifyIdx < workSizePerDataParityBlock; ++verifyIdx) {
            if (*(codeHost + verifyIdx) != *(outHost + verifyIdx)) {
                printf("not equal at mIdx:%d verifyIdx:%d\n", mIdx, verifyIdx);
                break;
            }
        }
    }

    free(dataHost);
    free(codeHost);
    
    free(dataPtrs);
    free(codePtrs);
    
    free(bitmatrix);
    
    free(inHost);
    free(outHost);
    
    free_cuda_device_memory(devBitmatrix);
    free_cuda_device_memory(in);
    free_cuda_device_memory(out);
}

void measure_base_work_size_time_avg(int k, int m, int w,
                                     int workSizePerDataParityBlock, int loopNum,
                                     double *encodeTimeElapse, double *encodeTimeVariance,
                                     double *decodeTimeElapse, double *decodeTimeVariance){
    cudaSetDevice(DEVICE_SELCTED_ID);
    
    int kIdx, mIdx;
    int *bitmatrix = perasure_create_bitmatrix(k, m, w), *devBitmatrix, *devDecodingBitmatrix;
    int loopIdx;
    int threadDimx = THREADS_PER_BLOCK;
    int packetsPerBlock = threadDimx/w *w;
    int blockNum;
    int workSizeInLong;
    
    char *dataHost, *codeHost;
    char *inHost, *outHost;
    
    char *data, *code;
    char *in, *out;
    
    double *encodingTimeArray;
    double *decodingTimeArray;

    
    workSizePerDataParityBlock =  (int)align_value(workSizePerDataParityBlock, sizeof(long)*w);
    workSizeInLong = workSizePerDataParityBlock / sizeof(long);
    blockNum = (workSizeInLong/packetsPerBlock);
    
    if ((workSizeInLong%packetsPerBlock) != 0) {
        ++blockNum;
    }
    
    alloc_cuda_device_memory((void **)&(devBitmatrix), sizeof(int)*k*m*w*w);
    alloc_cuda_device_memory((void **)&(devDecodingBitmatrix), sizeof(int)*k*m*w*w);
    
    alloc_cuda_host_memory((void **)&(dataHost), workSizePerDataParityBlock*k);
    alloc_cuda_host_memory((void **)&(codeHost), workSizePerDataParityBlock*m);
    
    alloc_cuda_host_memory((void **)&(inHost), workSizePerDataParityBlock*k);
    alloc_cuda_host_memory((void **)&(outHost), workSizePerDataParityBlock*m);
    
    alloc_cuda_device_memory((void **)&data, workSizePerDataParityBlock*k);
    alloc_cuda_device_memory((void **)&code, workSizePerDataParityBlock*m);
    
    alloc_cuda_device_memory((void **)&in, workSizePerDataParityBlock*k);
    alloc_cuda_device_memory((void **)&out, workSizePerDataParityBlock*m);
    
    generateRandomValue(dataHost, workSizePerDataParityBlock*k);
    
    encodingTimeArray = talloc(double, loopNum);
    decodingTimeArray = talloc(double, loopNum);

    cudaMemcpy(devBitmatrix, bitmatrix, sizeof(int)*k*m*w*w, cudaMemcpyHostToDevice);
    cudaMemcpy(data, dataHost, workSizePerDataParityBlock*k, cudaMemcpyHostToDevice);
    
//    printf("warm up\n");
    //warm up
    measure_coding(data, code, devBitmatrix, k, m, w, threadDimx, blockNum, workSizeInLong);
    measure_coding(data, code, devBitmatrix, k, m, w, threadDimx, blockNum, workSizeInLong);
    measure_coding(data, code, devBitmatrix, k, m, w, threadDimx, blockNum, workSizeInLong);
    measure_coding(data, code, devBitmatrix, k, m, w, threadDimx, blockNum, workSizeInLong);

//    printf("start loop\n");
    for (loopIdx = 0; loopIdx < loopNum; ++loopIdx) {
        
        encodingTimeArray[loopIdx] = (double) measure_coding(data, code, devBitmatrix, k, m, w, threadDimx, blockNum, workSizeInLong);
        
        cudaMemcpy(codeHost, code, workSizePerDataParityBlock*m, cudaMemcpyDeviceToHost);

        int *erasures = talloc(int, k+m+1);
        int erasureNum = randomErasureM(erasures, m, k+m);
        int *erased = perasure_erasures_to_erased(k, m, erasures);
        int *matIdx = talloc(int, erasureNum);
        int *dm_ids = talloc(int, k);
        
        int *decodingBitmatrix = perasure_create_decoding_bitmatrix(k, m, w, bitmatrix,
                                                                matIdx, erasures, dm_ids);
        
        cudaMemcpy(devDecodingBitmatrix, decodingBitmatrix, sizeof(int)*k*erasureNum*w*w, cudaMemcpyHostToDevice);

        int datablockMissing = 0;

        for (kIdx = 0; kIdx < k; ++kIdx) {
//            printf("erased[%d]:%d\n",kIdx, erased[kIdx]);
            if (erased[kIdx] == 1) {
                datablockMissing = 1;
            }
        }
        
        
        if (datablockMissing == 0) {
            for (kIdx = 0; kIdx < k; ++kIdx) {
//                memcpy(inHost + kIdx * workSizePerDataParityBlock, dataHost + kIdx * workSizePerDataParityBlock, workSizePerDataParityBlock);
                cudaMemcpy(inHost + kIdx * workSizePerDataParityBlock, dataHost + kIdx * workSizePerDataParityBlock, workSizePerDataParityBlock, cudaMemcpyHostToHost);

            }
        }else{
            for (kIdx = 0; kIdx < k; ++kIdx) {
                if (dm_ids[kIdx] < k) {
//                    memcpy((inHost + kIdx * workSizePerDataParityBlock), (dataHost + dm_ids[kIdx] * workSizePerDataParityBlock), workSizePerDataParityBlock);
                    cudaMemcpy(inHost + kIdx * workSizePerDataParityBlock, (dataHost + dm_ids[kIdx] * workSizePerDataParityBlock), workSizePerDataParityBlock, cudaMemcpyHostToHost);

                }else{
//                    memcpy((inHost+ kIdx * workSizePerDataParityBlock), (codeHost + (dm_ids[kIdx] - k) * workSizePerDataParityBlock), workSizePerDataParityBlock);
                    cudaMemcpy(inHost + kIdx * workSizePerDataParityBlock, (codeHost + (dm_ids[kIdx] - k) * workSizePerDataParityBlock), workSizePerDataParityBlock, cudaMemcpyHostToHost);
                }
            }
        }
        

        cudaMemcpy(in, inHost, workSizePerDataParityBlock*k, cudaMemcpyHostToDevice);

        decodingTimeArray[loopIdx] = measure_coding(in, out, devDecodingBitmatrix, k, erasureNum, w, threadDimx, blockNum, workSizeInLong);
//        printf("decode time:%fms\n", decodingTimeArray[loopIdx]);
        
        cudaMemcpy(outHost, out, workSizePerDataParityBlock*m, cudaMemcpyDeviceToHost);

        // verify
        int sIdx;
        int verifyWorkSize = workSizePerDataParityBlock;
        for (mIdx = 0; mIdx < erasureNum; ++mIdx) {
            char *src_ptr;
            
            if (matIdx[mIdx] < k) {
                src_ptr = dataHost + matIdx[mIdx] * workSizePerDataParityBlock;
            }else{
                src_ptr = codeHost + (matIdx[mIdx] - k) * workSizePerDataParityBlock;
            }
            
            
            for (sIdx = 0; sIdx < verifyWorkSize ; sIdx = sIdx + 1) {
                if (*(outHost + mIdx * workSizePerDataParityBlock + sIdx) != *(src_ptr + sIdx)) {
                    printf("matIdx[mIdx]: %d, %d %d \n",matIdx[mIdx], mIdx, sIdx);
                    break;
                }
            }
        }
        
        free(erasures);
        free(erased);
        free(matIdx);
        free(dm_ids);
        free(decodingBitmatrix);
    }
    
    calMeanVariance(encodingTimeArray, loopNum, encodeTimeElapse, encodeTimeVariance);
    calMeanVariance(decodingTimeArray, loopNum, decodeTimeElapse, decodeTimeVariance);
    
    free(bitmatrix);
    free(encodingTimeArray);
    free(decodingTimeArray);
    
    free_cuda_device_memory(devBitmatrix);
    free_cuda_device_memory(devDecodingBitmatrix);
    
    free_cuda_host_memory(dataHost);
    free_cuda_host_memory(codeHost);
    free_cuda_host_memory(inHost);
    free_cuda_host_memory(outHost);
    
    free_cuda_device_memory(data);
    free_cuda_device_memory(code);
    free_cuda_device_memory(in);
    free_cuda_device_memory(out);
    
    cudaDeviceReset();
}























